from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import inspect
import random
from itertools import product


def populate_kwargs(self, properties, kwargs):
    ''' Fills self based on keywords in properties '''
    for k, v in properties.items():
        value = kwargs[k] if k in kwargs else v
        if not hasattr(self, k):
            setattr(self, k, value)


def all_classes_of(mod):
    return inspect.getmembers(mod, lambda member: inspect.isclass(member) and
                              mod.__name__ in member.__module__)


class AbsoluteLocationUtils(object):
    @staticmethod
    def _get_abs_loc_features(game):
        fts = set()
        max_w, max_h = game.get_max_bounds()
        for x, y in product(range(max_w), range(max_h)):
            fts.add(AbsoluteLocationUtils._coords2loc(x, y))
        return sorted(list(fts))

    @staticmethod
    def _coords2loc(x, y):
        return "{0}x{1}y".format(x, y)


class MazeException(Exception):
    ''' Exception in creating maze, may try again after catching '''
    pass


def choice(lst):
    ''' So mazebase can catch these exceptions '''
    if len(lst) == 0:
        raise MazeException("Running choice on empty list")
    return random.choice(lst)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from random import choice
import sys

from mazebase.utils import mazeutils
import mazebase.items as mi


class Agent(mi.MazeItem):
    '''
    Agents are special items that can perform actions. We use a mix-ins model
    to specify Agent traits. To combine traits, simply subclass both
    Agent classes:

    # This agent can move and drop bread crumbs
    class SingleGoalAgent(mi.SingleTileMovable, mi.BreadcrumbDropping):
        pass

    To make a new agent trait, create the class, subclass from Agent, create
    the actions, and call self._add_action('id', self.__action)

    IMPORTANT: Any attributes defined outside of this module will not be
    featurized. Agents are featurized as a list of what they can 'do'
    '''
    __properties = dict(
        # Speed allows some agents to move faster than others
        speed=1
    )

    def __init__(self, **kwargs):
        mazeutils.populate_kwargs(self, self.__class__.__properties, kwargs)
        super(Agent, self).__init__(**kwargs)

        self.actions = {'pass': self._pass}
        self.PRIO = 100
        self._all_agents = [x[1] for x in
                            mazeutils.all_classes_of(sys.modules[__name__])]

    def _pass(self):
        pass

    def _add_action(self, id, func):
        assert id not in self.actions, "Duplicate action id"
        self.actions[id] = func

    def featurize(self):
        features = list(set(self.__get_all_superclasses(self.__class__)))
        return features

    def __get_all_superclasses(self, cls):
        all_superclasses = []
        for superclass in cls.__bases__:
            if superclass in self._all_agents:
                all_superclasses.append(superclass.__name__)
            all_superclasses.extend(self.__get_all_superclasses(superclass))
        return all_superclasses

    def _get_display_symbol(self):
        return (u' A ', None, None, None)


class NPC(Agent):
    ''' NPC Agents cannot be controlled by the player and moves randomly '''
    def get_npc_action(self):
        return (self.id, choice(self.actions))


class SingleTileMovable(Agent):
    ''' Can move up, down, left, and right 1 tile per turn '''
    def __init__(self, **kwargs):
        super(SingleTileMovable, self).__init__(**kwargs)

        self._add_action("up", self.__up)
        self._add_action("down", self.__down)
        self._add_action("left", self.__left)
        self._add_action("right", self.__right)

    def __dmove(self, dx, dy):
        x, y = self.location
        nloc = x + dx, y + dy
        # Cannot walk into blocks, agents, or closed doors
        if (self.game._tile_get_block(nloc, mi.Block) is None and
                self.game._tile_get_block(nloc, Agent) is None and
                (not self.game._tile_get_block(nloc, mi.Door) or
                 self.game._tile_get_block(nloc, mi.Door).isopen)):
            self.game._move_item(self.id, location=nloc)

    def __up(self):
        self.__dmove(0, 1)

    def __down(self):
        self.__dmove(0, -1)

    def __left(self):
        self.__dmove(-1, 0)

    def __right(self):
        self.__dmove(1, 0)


class BreadcrumbDropping(Agent):
    ''' Can drop breadcrumbs as an action '''
    def __init__(self, **kwargs):
        super(BreadcrumbDropping, self).__init__(**kwargs)
        self._add_action("breadcrumb", self.__drop_crumb)

    def __drop_crumb(self):
        if self.game._tile_get_block(self.location, mi.Breadcrumb) is None:
            self.game._add_item(mi.Breadcrumb(location=self.location))


class Pushing(Agent):
    '''
    Can push in the 4 cardinal directions. Pushing moves Pushable objects
    in one of four directions if there's no collision.
    '''
    def __init__(self, **kwargs):
        super(Pushing, self).__init__(**kwargs)
        self._add_action("push_up", self.__push_up)
        self._add_action("push_down", self.__push_down)
        self._add_action("push_left", self.__push_left)
        self._add_action("push_right", self.__push_right)

    def __dpush(self, dx, dy):
        x, y = self.location
        tx, ty = x + dx, y + dy
        nx, ny = tx + dx, ty + dy

        # Cannot push into other blocks or agents
        block = self.game._tile_get_block((tx, ty), mi.Pushable)
        if (block is not None and
                self.game._tile_get_block((nx, ny), Agent) is None and
                self.game._tile_get_block((nx, ny), mi.Block) is None):
            self.game._move_item(block.id, location=(nx, ny))

    def __push_up(self):
        self.__dpush(0, 1)

    def __push_down(self):
        self.__dpush(0, -1)

    def __push_left(self):
        self.__dpush(-1, 0)

    def __push_right(self):
        self.__dpush(1, 0)


class Toggling(Agent):
    ''' Can toggle on current space '''
    def __init__(self, **kwargs):
        super(Toggling, self).__init__(**kwargs)
        self._add_action("toggle_switch", self.__toggle)

    def __toggle(self):
        x, y = self.location
        switch = self.game._tile_get_block((x, y), mi.Switch)
        if switch is not None:
            switch.toggle()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from mazebase.items import MazeItem


class HasStatesMixin(object):
    _MAX_STATES = 10
    STATE_FEATURE = ["state{0}".format(i) for i in range(_MAX_STATES)]

    @classmethod
    def all_features(cls):
        return super(HasStatesMixin, cls).all_features() + cls.STATE_FEATURE


class Block(MazeItem):
    def __init__(self, **kwargs):
        super(Block, self).__init__(passable=False, **kwargs)

    def _get_display_symbol(self):
        return (None, None, 'on_white', None)


class Water(MazeItem):
    def __init__(self, **kwargs):
        super(Water, self).__init__(**kwargs)
        self.PRIO = -100

    def _get_display_symbol(self):
        return (None, None, 'on_blue', None)


class Corner(MazeItem):
    def __init__(self, **kwargs):
        super(Corner, self).__init__(**kwargs)

    def _get_display_symbol(self):
        return (u'   ', None, None, None)


class Goal(MazeItem):
    __MAX_GOAL_IDS = 10

    def __init__(self, id=0, **kwargs):
        super(Goal, self).__init__(**kwargs)
        self.goal_id = id
        assert self.goal_id < self.__MAX_GOAL_IDS,\
            "cannot create goal with id >{0}".format(
                self.__MAX_GOAL_IDS)

    def _get_display_symbol(self):
        return (u'*{0}*'.format(self.goal_id), 'red', None, None)

    def featurize(self):
        return super(Goal, self).featurize() +\
            ["goal_id" + str(self.goal_id)]

    @classmethod
    def all_features(cls):
        return super(Goal, cls).all_features() +\
            ["goal_id" + str(k) for k in range(cls.__MAX_GOAL_IDS)]


class Breadcrumb(MazeItem):
    def __init__(self, **kwargs):
        super(Breadcrumb, self).__init__(**kwargs)
        self.PRIO = -50

    def _get_display_symbol(self):
        return (u' . ', None, None, None)


class Pushable(Block):
    def __init__(self, **kwargs):
        super(Pushable, self).__init__(**kwargs)

    def _get_display_symbol(self):
        return (None, None, 'on_green', None)


class Switch(HasStatesMixin, MazeItem):
    def __init__(self, start_state=0, nstates=2, **kwargs):
        super(Switch, self).__init__(**kwargs)
        self.state = start_state
        self.nstates = nstates
        assert self.nstates < HasStatesMixin._MAX_STATES,\
            "cannot create switches with >{0} states".format(
                self.__MAX_SWITCH_STATES)

    def _get_display_symbol(self):
        return (str(self.state).rjust(3), 'cyan', None, None)

    def toggle(self):
        self.state = (self.state + 1) % self.nstates

    def featurize(self):
        return super(Switch, self).featurize() +\
            [self.STATE_FEATURE[self.state]]


class Door(HasStatesMixin, MazeItem):
    def __init__(self, open=False, state=0, **kwargs):
        super(Door, self).__init__(**kwargs)
        self.isopen = open
        self.state = state

    def _get_display_symbol(self):
        return (None if self.isopen else u'\u2588{0}\u2588'.format(self.state),
                None, None, None)

    def open(self):
        self.isopen = True

    def close(self):
        self.isopen = False

    def toggle(self):
        self.isopen = not self.isopen

    def featurize(self):
        return super(Door, self).featurize() + \
            ["open" if self.open else "closed", self.STATE_FEATURE[self.state]]

    @classmethod
    def all_features(cls):
        return super(Door, cls).all_features() +\
            ["open", "closed"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from random import randint

from mazebase.games import (
    WithWaterAndBlocksMixin,
    RewardOnEndMixin,
    BaseVocabulary,
)
from mazebase.utils import creationutils
from mazebase.utils.mazeutils import populate_kwargs, MazeException, choice
from mazebase.items import agents
import mazebase.items as mi


# So we don't have to type as much
class PBSType(RewardOnEndMixin, WithWaterAndBlocksMixin, BaseVocabulary):
    pass


# Agents we only use once don't have to be put in items/Agent.py :)
class PushBlockAgent(agents.SingleTileMovable, agents.Pushing):
    pass


class SwitchesAgent(agents.SingleTileMovable, agents.Toggling):
    pass


def _est(game, s, e):
    '''shorthand to estimate reward for an agent to move from s to e'''
    visited, path = creationutils.dijkstra(
        game, s, creationutils.agent_movefunc, True)
    return -visited[e]  # Returns distance, which is negation of reward


def pbwps(p, start, end):
    path = [end]
    while path[-1] != start:
        path.append(p[path[-1]])
    waypoints = []
    path.reverse()
    for i, j in zip(path[:-1], path[1:]):
        x, y = i
        nx, ny = j
        waypoints.append((2 * x - nx, 2 * y - ny))
    return waypoints


class PushBlock(PBSType):
    ''' Agent must push a block onto a switch '''

    def _reset(self):
        super(PushBlock, self)._reset()

        self.sw_loc = choice(creationutils.empty_locations(self))
        self.sw = mi.Switch(location=self.sw_loc)
        self._add_item(self.sw)

        x, y = choice(creationutils.empty_locations(self))
        self.pushable = mi.Pushable(location=(x, y))
        self._add_item(self.pushable)

        visited, p = creationutils.dijkstra(self, (x, y),
                                            creationutils.pushblock_movefunc)
        if self.sw_loc not in visited:
            raise MazeException("No path to sw")
        self.waypoints = pbwps(p, self.pushable.location, self.sw.location)

        x, y = choice(creationutils.empty_locations(self,
                                                    bad_blocks=[mi.Block]))
        self.agent = PushBlockAgent(location=(x, y))
        self._add_agent(self.agent, "PushBlockAgent")
        visited, _ = creationutils.dijkstra(self, (x, y),
                                            creationutils.agent_movefunc)
        if self.waypoints[0] not in visited:
            raise MazeException("No path to pushblock")

    def _side_information(self):
        return super(PushBlock, self)._side_information() + \
            [[self.FEATURE.PUSH] +
             self.pushable.featurize() +
             self.sw.featurize()
             ]

    def _finished(self):
        return self.pushable.location == self.sw_loc

    def _calculate_approximate_reward(self):
        '''Approximation used: Agent can move through push-block '''
        cur = self.agent.location
        r = 0
        for loc in self.waypoints:
            r += _est(self, cur, loc)
            cur = loc
        return super(PushBlock, self)._calculate_approximate_reward() + r


class PushBlockCardinal(PBSType):
    ''' Agent must push a block onto one of the 4 cardinal walls '''

    def _reset(self):
        super(PushBlockCardinal, self)._reset()

        x, y = choice(creationutils.empty_locations(self))
        self.pushable = mi.Pushable(location=(x, y))
        self._add_item(self.pushable)
        self.direction = choice([self.FEATURE.UP,
                                 self.FEATURE.DOWN,
                                 self.FEATURE.LEFT,
                                 self.FEATURE.RIGHT,
                                 ])
        if self.direction == self.FEATURE.UP:
            self.goals = set((i, self.height) for i in range(self.width))
        elif self.direction == self.FEATURE.DOWN:
            self.goals = set((i, 0) for i in range(self.width))
        elif self.direction == self.FEATURE.LEFT:
            self.goals = set((0, i) for i in range(self.height))
        elif self.direction == self.FEATURE.RIGHT:
            self.goals = set((self.width, i) for i in range(self.height))

        visited, p = creationutils.dijkstra(self, (x, y),
                                            creationutils.pushblock_movefunc)
        if not any(x in visited for x in self.goals):
            raise MazeException("No path to goal")
        closest = min(self.goals, key=lambda loc: visited[loc])
        self.waypoints = pbwps(p, self.pushable.location, closest)

        x, y = choice(creationutils.empty_locations(self,
                                                    bad_blocks=[mi.Block]))
        self.agent = PushBlockAgent(location=(x, y))
        self._add_agent(self.agent, "PushBlockCardinalAgent")

    def _side_information(self):
        return super(PushBlockCardinal, self)._side_information() + \
            [[self.FEATURE.PUSH, self.direction] + self.pushable.featurize()]

    def _finished(self):
        return any(x == self.pushable.location for x in self.goals)

    def _calculate_approximate_reward(self):
        '''
        Approximation used:
            Agent can move through push-block
            The point on the wall to move block to is the one closest to block
        '''
        cur = self.agent.location
        r = 0
        for loc in self.waypoints:
            r += _est(self, cur, loc)
            cur = loc
        return super(PushBlockCardinal,
                     self)._calculate_approximate_reward() + r


class Switches(PBSType):
    ''' Agent must toogle all switches to the same color '''

    __properties = dict(
        n_switches=2,
        switch_states=2,
    )

    def __init__(self, **kwargs):
        populate_kwargs(self, self.__class__.__properties, kwargs)
        super(Switches, self).__init__(**kwargs)

    def _reset(self):
        super(Switches, self)._reset()

        loc = choice(creationutils.empty_locations(self, bad_blocks=[mi.Block]))
        self.agent = SwitchesAgent(location=loc)
        self._add_agent(self.agent, "SwitchesAgent")

        visited, _ = creationutils.dijkstra(self, loc,
                                            creationutils.agent_movefunc)

        self._switches = []
        for _ in range(self.n_switches):
            loc = choice(creationutils.empty_locations(self))
            self._switches.append(mi.Switch(
                location=loc,
                nstates=self.switch_states,
                start_state=choice(range(self.switch_states)),
            ))
            self._add_item(self._switches[-1])
            if loc not in visited:
                raise MazeException("No path to goal")

    def _finished(self):
        return len(set(x.state for x in self._switches)) == 1

    def _side_information(self):
        return super(Switches, self)._side_information() + \
            [[self.FEATURE.SWITCH,
              self.FEATURE.STATE,
              self.FEATURE.SAME,
              ]]

    def _calculate_approximate_reward(self):
        '''Greedy solution that visits each switch in turn'''
        best = 1e100
        sw_colors = [sw.state for sw in self._switches]
        print(sw_colors)
        for i, sw in enumerate(self._switches):
            tmp = [(sw.state - c) % self.switch_states for c in sw_colors]
            # Heuristic for perferring not needing to flip switches
            best = min(best, sum(x if x > 0 else -2 for x in tmp))

        to_visit = [sw.location for sw in self._switches]
        loc = self.agent.location
        r = 0
        for i, sw in enumerate(self._switches):
            visited, path = creationutils.dijkstra(
                self, loc, creationutils.agent_movefunc, True)
            ind, loc = min(enumerate(to_visit), key=lambda x: visited[x[1]])
            r -= visited[loc]  # Reward is negative of path
            to_visit.remove(loc)
        return super(Switches,
                     self)._calculate_approximate_reward() + \
                r - best * self.turn_penalty


def add_vertical_wall(game):
    size = (game.width, game.height)
    dim = choice([0, 1])
    line = randint(1, size[1 - dim] - 2)
    opening = randint(0, size[dim] - 1)
    for i in range(size[dim]):
        if i != opening:
            loc = [line, line]
            loc[dim] = i
            game._add_item(mi.Block(location=loc))

    loc = [line, line]
    loc[dim] = opening

    return loc, dim


class LightKey(PBSType):
    ''' Agent must open a door with a switch and go to the goal '''
    __properties = dict(
        switch_states=2,
    )

    def __init__(self, **kwargs):
        populate_kwargs(self, self.__class__.__properties, kwargs)
        super(LightKey, self).__init__(**kwargs)

    def _reset(self):
        hole, dim = add_vertical_wall(self)

        # Add the door
        self.door = mi.Door(location=hole,
                            state=choice(range(1, self.switch_states)))
        self._add_item(self.door)

        # Add additional blocks and waters
        super(LightKey, self)._reset()

        # Add the goal, agent, and switch
        loc = choice(creationutils.empty_locations(
            self, bad_blocks=[mi.Block, mi.Door]))
        self.goal = mi.Goal(location=loc)
        self._add_item(self.goal)
        side = choice([-1, 1])

        def mask_func(x, y):
            return side * ((x, y)[1 - dim] - hole[1 - dim]) > 0

        loc = choice(creationutils.empty_locations(
            self, bad_blocks=[mi.Block, mi.Door, mi.Goal], mask=mask_func))
        self.sw = mi.Switch(location=loc, nstates=self.switch_states)
        self._add_item(self.sw)

        loc = choice(creationutils.empty_locations(
            self, bad_blocks=[mi.Block, mi.Door], mask=mask_func))
        self.agent = SwitchesAgent(location=loc)
        self._add_agent(self.agent, "LightKeyAgent")

        visited, _ = creationutils.dijkstra(self, loc,
                                            creationutils.agent_movefunc)
        if self.goal.location not in visited or self.sw.location not in visited:
            raise MazeException("No path to goal")

    def _step(self):
        # Hook the door up to the switch
        if self.sw.state == self.door.state:
            self.door.open()
        else:
            self.door.close()

    def _finished(self):
        return self.agent.location == self.goal.location

    def _side_information(self):
        return super(LightKey, self)._side_information() + \
            [[self.FEATURE.GOTO] + self.goal.featurize()]

    def _calculate_approximate_reward(self):
        x, y = self.door.location
        saved = self._approx_reward_map[x][y]
        self._approx_reward_map[x][y] = -1e100
        r = _est(self, self.agent.location, self.goal.location)
        if r < -1e90:
            self._approx_reward_map[x][y] = saved
            r = _est(self, self.agent.location, self.sw.location) + \
                _est(self, self.sw.location, self.goal.location)
        return super(LightKey, self)._calculate_approximate_reward() + r


class BlockedDoor(PBSType):
    ''' Agent must push a block out of the way and reach a goal '''

    def _reset(self):
        hole, dim = add_vertical_wall(self)

        # Add the pushblock
        self.pushable = mi.Pushable(location=hole)
        self._add_item(self.pushable)

        # Add additional blocks and waters
        super(BlockedDoor, self)._reset()

        # Add the goal, and agent
        loc = choice(creationutils.empty_locations(
            self, bad_blocks=[mi.Block, mi.Door]))
        self.goal = mi.Goal(location=loc)
        self._add_item(self.goal)

        loc = choice(creationutils.empty_locations(
            self, bad_blocks=[mi.Block, mi.Door]))
        self.agent = PushBlockAgent(location=loc)
        self._add_agent(self.agent, "BlockedDoorAgent")

        self._remove_item(self.pushable.id)
        visited, _ = creationutils.dijkstra(self, loc,
                                            creationutils.agent_movefunc)
        if self.goal.location not in visited:
            raise MazeException("No path to goal")
        self._add_item(self.pushable)

    def _finished(self):
        return self.agent.location == self.goal.location

    def _side_information(self):
        return super(BlockedDoor, self)._side_information() + \
            [[self.FEATURE.GOTO] + self.goal.featurize()]

    def _calculate_approximate_reward(self):
        r = _est(self, self.agent.location, self.goal.location)
        if r < -1e90:
            self._remove_item(self.pushable.id)
            r = _est(self, self.agent.location, self.goal.location)
            r -= 4 * self.turn_penalty  # Heuristic for pushing block
            self._add_item(self.pushable)
        return super(BlockedDoor, self)._calculate_approximate_reward() + r



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import abc
import itertools
import six
from mazebase.utils.mazeutils import AbsoluteLocationUtils

MAX_SENTENCE_SIZE = 10


########################
# Abstract classes
########################

@six.add_metaclass(abc.ABCMeta)
class Featurizer(object):

    @abc.abstractmethod
    def featurize(self, game, id):
        pass

    @abc.abstractmethod
    def all_possible_features(self, game):
        '''Extra features added by the featurizer'''
        return ['']


@six.add_metaclass(abc.ABCMeta)
class BaseGridFeaturizer(Featurizer):
    def __init__(self, *args, **kwargs):
        self.max_info_length = kwargs.pop('max_info_length', 10)
        self.max_infos = kwargs.pop('max_infos', 10)

    def featurize(self, game, id):
        features = (self._featurize_grid(game, id),
                    self._featurize_side_info(game, id))
        return features

    def _featurize_side_info(self, game, id):
        features = game._side_info()
        if len(features) > self.max_info_length:
            raise Exception("Too much side info too long to featurize")
        features += [[] for i in range(self.max_infos - len(features))]
        for feat in features:
            if len(feat) > self.max_info_length:
                raise Exception("Info feature too long")
            feat += [""] * (self.max_info_length - len(feat))
        return features

    @abc.abstractmethod
    def _featurize_grid(self, game, id):
        pass


@six.add_metaclass(abc.ABCMeta)
class SentenceFeaturizer(Featurizer):
    ''' Abstract class for sentence based featurizers '''
    def __init__(self, *args, **kwargs):
        self.max_sentence_length = kwargs.pop('max_sentence_length', 10)
        self.max_sentences = kwargs.pop('max_sentences', 100)

    def featurize(self, game, id):
        features = self._featurize(game, id) + game._side_info()
        if len(features) > self.max_sentences:
            raise Exception("Too many objects to featurize")
        # Do padding for features
        features += [[] for i in range(self.max_sentences - len(features))]
        for feat in features:
            if len(feat) > self.max_sentence_length:
                raise Exception("Sentence feature too long")
            feat += [""] * (self.max_sentence_length - len(feat))
        return features

    @abc.abstractmethod
    def _featurize(self, game, id):
        pass


class AbsoluteLocationMixin(AbsoluteLocationUtils):
    '''Featurizer uses absolute locations'''
    def all_possible_features(self, game):
        return self._get_abs_loc_features(game)


class RelativeLocationMixin(object):
    '''Featurizer uses relative locations.
    bounds specifies maximum 'sight' range'''
    def __init__(self, **kwargs):
        self.bounds = kwargs.pop('bounds', 5)
        super(RelativeLocationMixin, self).__init__(**kwargs)

    def all_possible_features(self, game):
        fts = set(super(RelativeLocationMixin, self)
                  .all_possible_features(game))
        for x, y in itertools.product(range(-self.bounds, self.bounds),
                                      range(-self.bounds, self.bounds)):
            fts.add(self._coords2loc(x, y))
        return list(sorted(fts))

    @staticmethod
    def _coords2loc(x, y):
        return "d{0}x{1}y".format(x, y)

########################
# Featurizers
########################


class GridFeaturizer(BaseGridFeaturizer):
    '''
    A list of featurizations of objects, in a matrix of width by height.
    Call game.all_possible_features to get a dictionary of all features.

    Returns:
        [
            grid_features: featurize the game world like a grid,
            side_info: game specific info blocks, in an ordered list
        ]
    '''

    def _featurize_grid(self, game, id):
        max_w, max_h = game.get_max_bounds()
        features = [[[] for y in range(max_w)]
                     for x in range(max_h)]
        for (x, y) in itertools.product(range(game.width), range(game.height)):
            itemlst = game._map[x][y]
            for item in itemlst:
                if not item.visible:
                    continue
                features[x][y] += item.featurize()

        return features

    def all_possible_features(self, game):
        return super(GridFeaturizer, self).all_possible_features(game)


class RelativeGridFeaturizer(AbsoluteLocationMixin, BaseGridFeaturizer):
    '''
    A list of featurizations of objects, in a matrix of width by height.
    Call game.all_possible_features to get a dictionary of all features.

    kwargs:
        bounds
        notify = add OUT_OF_BOUNDS features or not

    Returns:
        [
            grid_features: featurize the game world like a grid,
            side_info: game specific info blocks, in an ordered list
        ]
    '''
    def __init__(self, **kwargs):
        self.bounds = kwargs.pop('bounds', 5)
        self.notify = kwargs.pop('notify', False)
        super(RelativeGridFeaturizer, self).__init__(**kwargs)

    def _featurize_grid(self, game, id):
        tx, ty = game._items[id].location
        max_w, max_h = game.get_max_bounds()
        features = [[[] for y in range(2 * self.bounds - 1)]
                     for x in range(2 * self.bounds - 1)]
        center = self.bounds - 1
        features[center][center].append(self._coords2loc(tx, ty))
        for (x, y) in itertools.product(range(2 * self.bounds - 1),
                                        range(2 * self.bounds - 1)):
            nx, ny = tx + x - center, ty + y - center
            if not (0 <= nx < game.width and 0 <= ny < game.height):
                if self.notify: features[x][y].append("OUT_OF_BOUNDS")
                continue
            itemlst = game._map[nx][ny]
            for item in itemlst:
                if not item.visible:
                    continue
                features[x][y] += item.featurize()

        return features

    def all_possible_features(self, game):
        fts = super(RelativeGridFeaturizer, self).all_possible_features(game)
        fts += ["OUT_OF_BOUNDS"]
        return list(sorted(fts))


class SentenceFeaturesAbsolute(AbsoluteLocationMixin, SentenceFeaturizer):
    '''
    A list of featurizations of objects in the map. The objects are given as
    "sentences", with an absolute location in the map.

    kwargs:
        max_sentence_length
        max_sentences
    '''

    def __init__(self, **kwargs):
        super(SentenceFeaturesAbsolute, self).__init__(**kwargs)

    def _featurize(self, game, id):
        max_w, max_h = game.get_max_bounds()
        features = []
        for id, item in game._items.items():
            if not item.visible:
                continue
            feat = item.featurize()
            location_feature = self._coords2loc(*item.location)
            features.append([location_feature] + feat)

        return features


class SentenceFeaturesRelative(RelativeLocationMixin, SentenceFeaturizer):
    '''
    A list of featurizations of objects in the map. The objects are given as
    "sentences", with an absolute location in the map. You can initialize
    this with a parameter that decides how far the agent can see.

    kwargs:
        max_sentence_length
        max_sentences
        bounds
    '''

    def _featurize(self, game, id):
        tx, ty = game._items[id].location
        max_w, max_h = game.get_max_bounds()
        features = []
        for id, item in game._items.items():
            if not item.visible:
                continue
            x, y = item.location
            dx, dy = tx - x, ty - y
            if not (-self.bounds < dx < self.bounds and
                    -self.bounds < dy < self.bounds):
                continue
            feat = item.featurize()
            location_feature = self._coords2loc(dx, dy)
            features.append([location_feature] + feat)

        return features

##########################
# Utility functions
##########################


def vocabify(game, observation, np=None):
    '''
    Changes the outputs of SentenceFeaturizer subclasses to a numerical
    representation. Also works on side_info.

    Pass in numpy module to np to use it instead of lists.
    '''
    vocab = dict([(b, a) for a, b in
                  enumerate(game.all_possible_features())])
    if np is None:
        for sent in observation:
            for i, word in enumerate(sent):
                sent[i] = vocab[word]
        return observation
    else:
        shape = (len(observation), len(observation[0]))
        arr = np.zeros(shape)
        for x, y in itertools.product(range(shape[0]), range(shape[1])):
            arr[x][y] = vocab[observation[x][y]]
        return arr


def grid_one_hot(game, observation, np=None):
    '''
    THIS ISN'T ACTUALLY ONE HOT, IT IS FEW HOT

    In place transformation:
    Changes the outputs of GridFeaturizers to a few hot representation,
    with feature plane size `x \\times y \\times nfeatures`.

    pass in the numpy module to np to return a numpy array,
    which is far more efficient than using lists.
    '''
    vocab = dict([(b, a) for a, b in
                  enumerate(game.all_possible_features())])
    if np is None:
        for x, col in enumerate(observation):
            for y, lst in enumerate(col):
                features = [0 for w in vocab]
                for feat in lst:
                    features[vocab[feat]] = 1
                observation[x][y] = features
        return observation
    else:
        xm, ym, zm = len(observation[0]), len(observation[1]), len(vocab)
        arr = np.zeros((xm, ym, zm))
        for x, y in itertools.product(range(xm), range(ym)):
            for feat in observation[x][y]:
                arr[x][y][vocab[feat]] = 1
        return arr


def grid_one_hot_sparse(game, observation, np=None):
    '''
    No longer in place.

    Returns a sparse list of (x, y, vocab_i), but otherwise same as above.

    Mostly used for python-lua bridge, since the communication costs
    there are relatively high
    '''
    vocab = dict([(b, a) for a, b in
                  enumerate(game.all_possible_features())])
    res = []
    for x, col in enumerate(observation):
        for y, lst in enumerate(col):
            for feat in lst:
                res.append((x, y, vocab[feat]))
    return res



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from random import shuffle, random
from collections import defaultdict
import itertools
import mazebase.items as mi
from .mazeutils import MazeException


def sprinkle(game, tiles, tilemask=None):
    '''
    Sprinkles blocks into a map. Tiles is given in the format like:
        [(MazeItem, float), ...] ex.
        [(Block, .5)]
    where we sprinkle MazeItem with the percent chance given by the second arg.

    Defaults to generating on empty tiles, but you can override this with
    tilemask and specify a list of locations.

    Returns list of item ids
    '''
    if tilemask is None:
        tilemask = empty_locations(game)

    ids = []
    for (x, y) in tilemask:
        shuffle(tiles)
        for tile, pct in tiles:
            if random() < pct:
                ids.append(game._add_item(tile(location=(x, y))))
                break

    return ids


def empty_locations(game, bad_blocks=None, mask=lambda x, y: True):
    '''By default, finds empty locations in the map.
    If bad_blocks is not none, then finds locations without any bad_blocks,
    but maybe with other block types
    mask is a function that provides valid coordinates
    '''
    empties = []
    for x, y in itertools.product(range(game.width), range(game.height)):
        if not mask(x, y):
            continue
        itemlst = game._map[x][y]
        if bad_blocks is None and itemlst == []:
            empties.append((x, y))
        elif bad_blocks is not None and not any(
                isinstance(item, typ) for
                item, typ in itertools.product(itemlst, bad_blocks)):
            empties.append((x, y))
    return empties


def dijkstra(game, initial, movefunc, weighted=False):
    '''
    Accepts:
        game
        initial: (x, y) tuple of start location
        movefunc: f(loc) determines the locations you can move to from loc
        weighted: use the _approx_reward_map instead of # of moves

    Returns:
        visited: dictionary of {location: distance} pairs
        path: dictionary of {location: previous_location} pairs
    '''
    visited = defaultdict(lambda: 1e309)
    visited[initial] = 0
    path = {}

    nodes = set(itertools.product(range(game.width), range(game.height)))
    while nodes:
        current = nodes.intersection(visited.keys())
        if not current:
            break
        min_node = min(current, key=visited.get)
        nodes.remove(min_node)
        current_weight = visited[min_node]
        x, y = min_node

        for edge in movefunc(game, min_node):
            # Maximize reward by minimizing "distance = - reward"
            w = -game._approx_reward_map[edge[0]][edge[1]] if weighted else 1
            weight = current_weight + w
            if edge not in visited or weight < visited[edge]:
                visited[edge] = weight
                path[edge] = min_node

    return visited, path


def __movefunc_helper(game, loc, movefunc_helper):
    res = []
    x, y = loc
    for dx, dy in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
        nx, ny = x + dx, y + dy
        if not game._in_bounds((nx, ny)):
            continue
        if movefunc_helper(game, loc, (dx, dy)):
            res.append((nx, ny))
    return res


def agent_movefunc(game, loc):
    ''' Can move to non-block spaces '''
    def helper(game, loc, dloc):
        x, y = loc
        dx, dy = dloc
        nx, ny = x + dx, y + dy
        return game._tile_get_block((nx, ny), mi.Block) is None

    return __movefunc_helper(game, loc, helper)


def pushblock_movefunc(game, loc):
    ''' Can move if tile behind and in front are not blocked (so agent can push
    from behind) '''
    def helper(game, loc, dloc):
        x, y = loc
        dx, dy = dloc
        tx, ty = x - dx, y - dy
        nx, ny = x + dx, y + dy
        return (game._in_bounds((tx, ty)) and
                game._tile_get_block((nx, ny), mi.Block) is None and
                game._tile_get_block((tx, ty), mi.Block) is None)

    return __movefunc_helper(game, loc, helper)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from random import shuffle, randint
from six.moves import range
from itertools import product

from . import (
    BaseMazeGame,
    WithWaterAndBlocksMixin,
    RewardOnEndMixin,
    BaseVocabulary,
    AbsoluteLocationVocabulary,
)
from mazebase.utils import creationutils
from mazebase.utils.mazeutils import populate_kwargs, MazeException, choice
from mazebase.items import agents
import mazebase.items as mi


class MovingAgent(agents.SingleTileMovable):
    '''Agents we only use once don't have to be put in items/Agent.py :)'''
    pass


# So we don't have to type as much
class GoalType(RewardOnEndMixin, WithWaterAndBlocksMixin, BaseVocabulary):
    pass


def _est(game, s, e):
    '''shorthand to estimate reward for an agent to move from s to e'''
    visited, path = creationutils.dijkstra(
        game, s, creationutils.agent_movefunc, True)
    return -visited[e]  # Returns distance, which is negation of reward


class SingleGoalApproximateRewardMixin(BaseMazeGame):
    def _calculate_approximate_reward(self):
        r = _est(self, self.agent.location, self.goal.location)
        return super(SingleGoalApproximateRewardMixin,
                     self)._calculate_approximate_reward() + r


class SingleGoal(GoalType):
    ''' Agent moves to a single goal '''

    def _reset(self):
        super(SingleGoal, self)._reset()

        loc = choice(creationutils.empty_locations(self))
        self.goal = mi.Goal(location=loc)
        self._add_item(self.goal)

        loc = choice(creationutils.empty_locations(self, bad_blocks=[mi.Block]))
        self.agent = MovingAgent(location=loc)
        self._add_agent(self.agent, "SingleGoalAgent")

        visited, _ = creationutils.dijkstra(self, loc,
                                            creationutils.agent_movefunc)
        if self.goal.location not in visited:
            raise MazeException("No path to goal")

    def _side_information(self):
        return super(SingleGoal, self)._side_information() + \
            [[self.FEATURE.GOTO] + self.goal.featurize()]

    def _finished(self):
        return self.agent.location == self.goal.location


class MultiGoals(GoalType):
    ''' Agent must visit each goal in order, without penalty for visiting out
    of order '''

    __properties = dict(
        n_goals=3,
    )

    def __init__(self, **kwargs):
        populate_kwargs(self, self.__class__.__properties, kwargs)
        super(MultiGoals, self).__init__(**kwargs)

    def _reset(self):
        super(MultiGoals, self)._reset()

        self.goals = []
        for i in range(self.n_goals):
            x, y = choice(creationutils.empty_locations(self))
            self.goals.append(mi.Goal(location=(x, y), id=i))
            self._add_item(self.goals[i])
        shuffle(self.goals)
        self.v = 0

        x, y = choice(creationutils.empty_locations(self,
                                                    bad_blocks=[mi.Block]))
        self.agent = MovingAgent(location=(x, y))
        self._add_agent(self.agent, "MultiGoalsAgent")

        visited, _ = creationutils.dijkstra(self, (x, y),
                                            creationutils.agent_movefunc)
        if not all(goal.location in visited for goal in self.goals):
            raise MazeException("No path to goal")

    def _side_information(self):
        return super(MultiGoals, self)._side_information() + \
            [[self.FEATURE.ORDERED_OBJ[i], self.FEATURE.GOTO] + x.featurize()
             for i, x in enumerate(self.goals)]

    def _step(self):
        if self.agent.location == self.goals[self.v].location:
            self.v = self.v + 1

    def _finished(self):
        return self.v == self.n_goals

    def _calculate_approximate_reward(self):
        locs = [self.agent.location] + [g.location for g in self.goals]
        r = sum(_est(self, l1, l2) for l1, l2 in zip(locs[:-1], locs[1:]))
        return super(MultiGoals, self)._calculate_approximate_reward() + r


class TogglingAgent(agents.SingleTileMovable, agents.Toggling):
    pass


class ConditionedGoals(GoalType):
    ''' Agent must visit goals conditioned on the color of a switch. Penalty
    for visiting a wrong goal '''

    __properties = dict(
        n_goals=3,
        n_colors=3,
        goal_penalty=0.2
    )

    def __init__(self, **kwargs):
        populate_kwargs(self, self.__class__.__properties, kwargs)
        super(ConditionedGoals, self).__init__(**kwargs)

    def _reset(self):
        super(ConditionedGoals, self)._reset()

        x, y = choice(creationutils.empty_locations(self))
        self.sw = mi.Switch(
            location=(x, y),
            nstates=self.n_colors,
            start_state=choice(range(self.n_colors)),
        )
        self._add_item(self.sw)

        self.goals = []
        for i in range(self.n_goals):
            x, y = choice(creationutils.empty_locations(self))
            self.goals.append(mi.Goal(location=(x, y), id=i))
            self._add_item(self.goals[i])
        self.conditions = [randint(0, self.n_goals - 1) for _ in self.goals]

        x, y = choice(creationutils.empty_locations(self,
                                                    bad_blocks=[mi.Block]))
        self.agent = TogglingAgent(location=(x, y))
        self._add_agent(self.agent, "ConditionedGoalsAgent")

        visited, _ = creationutils.dijkstra(self, (x, y),
                                            creationutils.agent_movefunc)

        if (self.sw.location not in visited or
            not any(self.goals[i].location in visited
                    for i in set(self.conditions))):
            raise MazeException("No path to goal")

    def _get_reward(self, id):
        reward = super(ConditionedGoals, self)._get_reward(id)
        forbiddens = set(g.location for g in self.goals)
        if self.sw.state < len(self.conditions):
            forbiddens.remove(
                self.goals[self.conditions[self.sw.state]].location)
        if self.agent.location in forbiddens:
            reward -= self.goal_penalty

        return reward

    def _finished(self):
        if self.sw.state >= len(self.conditions):
            return False
        return self.agent.location == \
            self.goals[self.conditions[self.sw.state]].location

    def _side_information(self):
        # Featurize the goals, and add it to the features of some dummy
        # switches in the right state
        return super(ConditionedGoals, self)._side_information() + \
            [
                [self.FEATURE.IF] +
                mi.Switch(nstates=self.n_goals, start_state=i).featurize() +
                [self.FEATURE.GOTO] +
                self.goals[st].featurize()
            for i, st in enumerate(self.conditions)]

    def _calculate_approximate_reward(self):
        if self.sw.state >= len(self.conditions):
            best = -1e100
        else:
            best = _est(self, self.agent.location,
                        self.goals[self.conditions[self.sw.state]].location)
        sw_dist = _est(self, self.agent.location, self.sw.location)
        for i in range(min(self.n_goals, self.n_colors)):
            goal_loc = self.goals[self.conditions[i]].location
            best = max(
                best,
                _est(self, self.sw.location, goal_loc) -
                ((i - self.sw.state) % self.n_colors) * self.turn_penalty +
                sw_dist)
        return super(ConditionedGoals, self)._calculate_approximate_reward() +\
                best + self.goal_penalty

    def _accumulate_approximate_rewards(self):
        super(ConditionedGoals, self)._accumulate_approximate_rewards()
        for x, y in product(range(self.width), range(self.height)):
            if self._tile_get_block((x, y), mi.Goal) is not None:
                self._approx_reward_map[x][y] += -self.goal_penalty


class Exclusion(GoalType):
    ''' Agent must visit all goals except those stated in side_information '''

    __properties = dict(
        goal_penalty=0.5,  # penalty for stepping on excluded goal
        n_goals=3,
        visit_min=1,
        visit_max=-1,  # -1 = n_goals
    )

    def __init__(self, **kwargs):
        populate_kwargs(self, self.__class__.__properties, kwargs)
        super(Exclusion, self).__init__(**kwargs)

    def _reset(self):
        super(Exclusion, self)._reset()
        if self.visit_min < 1:
            raise MazeException("visit_min is not >= 1")
        if self.visit_max == -1:
            self.visit_max = self.n_goals
        self.visit = list(range(self.n_goals))
        shuffle(self.visit)
        to_visit = randint(self.visit_min, self.visit_max)
        self.exclude = self.visit[to_visit:]
        self.visit = self.visit[:to_visit]
        self.visit = dict((x, False) for x in self.visit)

        self.goals = []
        for i in range(self.n_goals):
            x, y = choice(creationutils.empty_locations(self))
            self.goals.append(mi.Goal(location=(x, y), id=i))
            self._add_item(self.goals[i])

        x, y = choice(creationutils.empty_locations(self,
                                                    bad_blocks=[mi.Block]))
        self.agent = TogglingAgent(location=(x, y))
        self._add_agent(self.agent, "ExclusionAgent")

        visited, _ = creationutils.dijkstra(self, (x, y),
                                            creationutils.agent_movefunc)

        if not all(goal.location in visited for goal in self.goals):
            raise MazeException("No path to goal")

    def _step(self):
        for i in self.visit.keys():
            if self.agent.location == self.goals[i].location:
                self.visit[i] = True

    def _finished(self):
        return all(self.visit.values())

    def _side_information(self):
        return super(Exclusion, self)._side_information() + \
            [[self.FEATURE.GOTO, self.FEATURE.ALL]] + \
            [[self.FEATURE.AVOID] +
             self.goals[i].featurize()
             for i in self.exclude]

    def _get_reward(self, id):
        reward = super(Exclusion, self)._get_reward(id)
        for i in self.exclude:
            if self.agent.location == self.goals[i].location:
                reward -= self.goal_penalty

        return reward

    def _calculate_approximate_reward(self):
        prev = self.agent.location
        visited = dict((x, False) for x in self.visit)
        so_far = 0
        for _ in self.visit:
            best_n = None
            best = -1e100
            for g in [g for g, b in visited.items() if not b]:
                t = _est(self, prev, self.goals[g].location)
                if t > best:
                    best = t
                    best_n = g
            prev = self.goals[best_n].location
            so_far += best
            visited[best_n] = True
        return super(Exclusion, self)._calculate_approximate_reward() + so_far

    def _accumulate_approximate_rewards(self):
        super(Exclusion, self)._accumulate_approximate_rewards()
        excludes = [self.goals[i].location for i in self.exclude]
        for x, y in product(range(self.width), range(self.height)):
            if (x, y) in excludes:
                self._approx_reward_map[x][y] += -self.goal_penalty


class GotoType(AbsoluteLocationVocabulary,
               GoalType,
               ):
    pass


class Goto(SingleGoalApproximateRewardMixin, GotoType):
    ''' Agent must go to an absolute location '''

    def _reset(self):
        super(Goto, self)._reset()

        loc = choice(creationutils.empty_locations(self))
        self.goal = mi.Goal(location=loc, visible=False)
        self._add_item(self.goal)

        loc = choice(creationutils.empty_locations(self,
                                                   bad_blocks=[mi.Block]))
        self.agent = MovingAgent(location=loc)
        self._add_agent(self.agent, "GotoAgent")

        visited, _ = creationutils.dijkstra(self, loc,
                                            creationutils.agent_movefunc)
        if self.goal.location not in visited:
            raise MazeException("No path to goal")

    def _side_information(self):
        return super(Goto, self)._side_information() + \
            [[self.FEATURE.GOTO, self._coords2loc(*self.goal.location)]]

    def _finished(self):
        return self.agent.location == self.goal.location


class GotoHidden(SingleGoalApproximateRewardMixin, GotoType):
    '''Agent must go to one of several goals, depending on side_information
    '''

    __properties = dict(
        n_goals=3,
    )

    def __init__(self, **kwargs):
        populate_kwargs(self, self.__class__.__properties, kwargs)
        super(GotoHidden, self).__init__(**kwargs)

    def _reset(self):
        super(GotoHidden, self)._reset()

        self.goals = []
        for i in range(self.n_goals):
            x, y = choice(creationutils.empty_locations(self))
            self.goals.append(mi.Goal(location=(x, y), id=i, visible=False))
            self._add_item(self.goals[i])

        self.goal = choice(self.goals)

        x, y = choice(creationutils.empty_locations(self,
                                                    bad_blocks=[mi.Block]))
        self.agent = MovingAgent(location=(x, y))
        self._add_agent(self.agent, "GotoHiddenAgent")

        visited, _ = creationutils.dijkstra(self, (x, y),
                                            creationutils.agent_movefunc)
        if self.goal.location not in visited:
            raise MazeException("No path to goal")

    def _side_information(self):
        return super(GotoHidden, self)._side_information() + \
            [[self._coords2loc(*goal.location)] + goal.featurize()
             for goal in self.goals] + \
            [[self.FEATURE.GOTO] + self.goal.featurize()]

    def _finished(self):
        return self.agent.location == self.goal.location



from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import abc
import copy
import random
import six
import types


def CurriculumWrappedGame(cls, curriculums=None, **kwargs):
    '''Introduces a curriculum by wrapping games

    Wraps a game and a set of keywords -> Curriculum mappings. By
    implementing
    custom curriculums, we can arbitrarily make games harder or easier.
    '''
    curriculums = curriculums or {}
    for kw, val in curriculums.items():
        kwargs[kw] = val.get()
    game = cls(**kwargs)

    def make_easier(self):
        random.choice(list(curriculums.values())).make_easier()

    def make_harder(self):
        random.choice(list(curriculums.values())).make_harder()

    def make_easiest(self):
        for cur in curriculums:
            cur.make_easiest()

    def make_hardest(self):
        for cur in curriculums:
            cur.make_hardest()

    max_bounds = game.get_max_bounds
    orig_reset = game.reset

    def get_max_bounds(self):
        if 'map_size' in curriculums:
            _, max_w, _, max_h = curriculums['map_size'].max
            return max_w, max_h
        return max_bounds()

    def reset(self):
        for kw, val in curriculums.items():
            setattr(self, kw, val.get())
        orig_reset()

    for func in [make_easier, make_harder, make_easiest, make_hardest,
                 get_max_bounds, reset]:
        setattr(game, func.__name__, types.MethodType(func, game))

    return game


@six.add_metaclass(abc.ABCMeta)
class CurriculumBase(object):
    '''
    self.current should store the current value of the curriculum to adjust
    '''

    def get(self):
        '''Get current value of curriculum'''
        return self.current

    @abc.abstractmethod
    def make_easier(self):
        pass

    @abc.abstractmethod
    def make_harder(self):
        pass

    @abc.abstractmethod
    def make_hardest(self):
        pass

    @abc.abstractmethod
    def make_easiest(self):
        pass


class NumericCurriculum(CurriculumBase):
    '''A curriculum on an integer variable

    Args:
        init: Where to start the curriculum
        min: minimum value to take
        max: maximum value to take
        step: harder and easier will adjust by step size
            this can be a callable function, if you want to generate steps
            randomly for example
    '''
    def __init__(self, init, min, max, step=1):
        assert min <= init <= max, "NumericCurriculum must be initialized"\
            "between min and max"
        self.min = min
        self.max = max
        self.step = step
        self.current = init
        if not callable(self.step):
            self.step = lambda: step

    def make_easier(self):
        self.current = max(self.current - self.step(), self.min)

    def make_harder(self):
        self.current = min(self.current + self.step(), self.max)

    def make_easiest(self):
        self.current = self.min

    def make_hardest(self):
        self.current = self.max


class MapSizeCurriculum(CurriculumBase):
    '''A curriculum on an integer variable

    Note:
        All arguments are of the form (min_x, max_x, min_y, max_y)
        Adjustment will try to move one of the values by 1, respecting that
        min will always be less than max. Sometimes, the adjustment will not
        move.

    Args:
        init: Where to start the curriculum
        min: minimum value to take
        max: maximum value to take
        step: harder and easier will adjust by step size
    '''
    def __init__(self, init, min, max):
        self.min = list(min)
        self.max = list(max)
        self.current = list(init)
        for init_v, max_v, min_v in zip(init, max, min):
            assert min_v <= init_v <= max_v, \
                "Initial sizes must be between min and max in MapSizeCurriculum"

    def __update(self, delta):
        ind = random.choice(range(4))
        self.current[ind] = self.current[ind] + delta

        # Enforce that current never moves above the min and max
        self.current[ind] = min(self.current[ind], self.max[ind])
        self.current[ind] = max(self.current[ind], self.min[ind])

        # Enforce that min never moves above max and vice versa
        if ind == 0:
            self.current[0] = min(self.current[0], self.current[1])
        if ind == 1:
            self.current[1] = max(self.current[0], self.current[1])
        if ind == 2:
            self.current[2] = min(self.current[2], self.current[3])
        if ind == 3:
            self.current[3] = max(self.current[2], self.current[3])

    def make_easier(self):
        self.__update(-1)

    def make_harder(self):
        self.__update(1)

    def make_easiest(self):
        self.current = copy.copy(self.min)

    def make_hardest(self):
        self.current = copy.copy(self.max)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import abc
import six
from copy import deepcopy

from mazebase.utils.mazeutils import populate_kwargs


@six.add_metaclass(abc.ABCMeta)
class MazeItem(object):
    '''
    Maze items should not share state
    PRIO determines priority of visibility when viewing the object,
        and has no effect on the game. A higher priority object is always
        diplayed first
    '''

    __properties = dict(
        location=(0, 0),
        visible=True,
        passable=True,
    )

    def __init__(self, **kwargs):
        populate_kwargs(self, self.__class__.__properties, kwargs)

        self.game = None
        self.PRIO = 0

    def _get_display_symbol(self):
        '''
        -> (text, foreground, background, attributes)

        text: should be 3 characters
        foreground: see termcolor.py
        background: see termcolor.py
        attributes: see termcolor.py
        '''
        return (None, None, None, None)

    def clone(self):
        return deepcopy(self)

    def featurize(self):
        ''' Return a list of the features for this item '''
        return [type(self).__name__]

    @classmethod
    def all_features(cls):
        '''
        All possible features for this item.
        Must implement if subclass implements featurize()
        '''
        return [cls.__name__]



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

requirements = ['six']

setup(
    name='mazebase',
    version='0.0.1',
    description='Game based library for reinforcement learning',
    long_description=long_description,
    url='https://github.com/facebook/mazebase',
    author='Nantas Nardelli',
    author_email='nantas@robots.ox.ac.uk',
    license='BSD',
    classifiers=[
        # TODO change when we are sure the library is solid
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='reinforcement-learning machine-learning artificial-intelligence',
    package_data={'mazebase': ['images/*.*'], },
    package_dir={'mazebase': 'py/mazebase'},
    packages=find_packages('py/'),
    install_requires=requirements,
    extras_require={}, )



from .mazegame import (
    MazeGame,
    BaseMazeGame,
    WithWaterAndBlocksMixin,
    RewardOnEndMixin,
    AbsoluteLocationVocabulary,
    BaseVocabulary,
)
from .goal_based_games import (
    SingleGoal,
    MultiGoals,
    ConditionedGoals,
    Exclusion,
    Goto,
    GotoHidden,
)
from .blocks_switches_games import (
    PushBlock,
    PushBlockCardinal,
    Switches,
    LightKey,
    BlockedDoor,
)



from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import abc
import logging
import random
import six
import uuid
from collections import OrderedDict
from itertools import chain, product

from mazebase.games import featurizers
from mazebase.termcolor import cprint
from mazebase.utils import creationutils
import mazebase.utils.mazeutils as mazeutils
import mazebase.items as mi
import mazebase.items.agents as agents


class MazeGame(object):
    '''
    Give it a list of games, and MazeGame will pick a random one for
    you to play. Check BaseMazeGame for list of usable functions. This
    secretly "inherits" from BaseMazeGame.
    '''

    def __init__(self, games,
                 featurizer=featurizers.SentenceFeaturesRelative(bounds=5)):
        self.featurizer = featurizer
        self.games = games

        # Get a union of the map sizes
        max_w, max_h = zip(*[game.get_max_bounds() for game in games])
        self.__max_bounds = (max(max_w), max(max_h))

        # Overwrite the featurizer with our current one
        for game in self.games:
            game._set_featurizer(self.featurizer)
        self.reset()

    def __getattr__(self, name):
        return getattr(self.game, name)

    def all_possible_features(self):
        feats = set()
        for game in self.games:
            feats.update(game.all_possible_features())
        return feats

    def all_possible_actions(self):
        actions = set()
        for game in self.games:
            actions.update(game.all_possible_actions())
        return list(sorted(actions))

    def get_max_bounds(self):
        return self.__max_bounds

    def reset(self):
        self.game = random.choice(self.games)
        self.game.reset()

    @classmethod
    def all_features(cls):
        return []


@six.add_metaclass(abc.ABCMeta)
class BaseMazeGame(object):
    '''
    Functions of interest to use - check doctrings for details:
        observe()   Returns the current observation
        is_over()   Whether the game is in a terminal state.
        act(action) Performs action, which must be in actions()
        reward()    Reward experienced by last action
        reward_so_far()     Reward during current episode
        approx_best_reward()    Approximation of optimal reward

        reset()     Gives a random game initialization
        display()   Simply prints a visualization of the game
        actions()   Currently allowed actions for current agent
        current_agent()     Returns current agent that is acting

    Some functions to do with game properties
        get_max_bounds()    Max game map size across randomizations
        all_possible_features()     Self descriptive
        all_possible_actions()  for all possible agents
        all_features()  All features for current game
        actions()   All possible actions for current agent
        current_agent()     Which agent will act

    List of functions to override for your own games:
        Required:
            _reset()
            _finished()
            _get_reward(agent_id)
            _side_information()  Game specific info features
        To support approximating reward:
            _accumulate_approximate_rewards()  Fills game._approx_reward_map
            _calculate_approximate_reward()  Called once per reset()
        Other functionality:
            _step()  hook called after every act().

    We use the root logger to log error messages, set the logger level
    to DEBUG to expose possible errors
    '''

    __properties = dict(
        featurizer=featurizers.SentenceFeaturesRelative(bounds=5),
        map_size=(5, 10, 5, 10),  # (min_x, max_x, min_y, max_y)
        turn_penalty=0.1,
    )

    def __init__(self, **kwargs):
        '''
        kwargs:
            featurizer: featurizer to use when doing observe()
            map_size: (x_min, x_max, y_min, y_max), draw uniformly and randomly
        '''
        mazeutils.populate_kwargs(self, self.__class__.__properties, kwargs)
        super(BaseMazeGame, self).__init__()
        self.game_name = uuid.uuid4().hex
        self.__all_possible_features = None
        self.__reward = 0
        self.reset()

    ####################
    # Utility functions
    ####################

    def display(self):
        ''' Displays the game map for visualization '''
        cprint(' ' * (self.width + 2) * 3, None, 'on_white')
        for y in reversed(range(self.height)):
            cprint('   ', None, 'on_white', end="")
            for x in range(self.width):
                itemlst = sorted(filter(lambda x: x.visible, self._map[x][y]),
                                 key=lambda x: x.PRIO)
                disp = [u'   ', None, None, None]
                for item in itemlst:
                    config = item._get_display_symbol()
                    for i, v in list(enumerate(config))[1:]:
                        if v is not None:
                            disp[i] = v
                    s = config[0]
                    if s is None:
                        continue
                    d = list(disp[0])
                    for i, char in enumerate(s):
                        if char != ' ':
                            d[i] = char
                    disp[0] = "".join(d)
                text, color, bg, attrs = disp
                cprint(text, color, bg, attrs, end="")
            cprint('   ', None, 'on_white')
        cprint(' ' * (self.width + 2) * 3, None, 'on_white')
        pass

    def observe(self):
        '''
        Returns:
            id: id of current agent to make an action
            observation: featurized version of map
        '''
        id = self.current_agent()
        return {
            'id': id,
            'reward': self.__reward,
            'observation': self._featurize(id),
        }

    def is_over(self):
        return self._finished()

    def reward(self):
        ''' Reward experienced by the last action taken. 0 if no action has
        been taken.'''
        return self.__reward

    def reward_so_far(self):
        return self.__reward_so_far

    def approx_best_reward(self):
        return self.__approx_best

    def reset(self):
        '''
        Wrapper to try 100 times, since sometimes the random generation
        screws up. Calls _reset to reset the map to a random initial state.
        Override _reset when creating a new game. Reset logic is in here
        so every subclass has access to reset variables correctly.
        '''
        for i in range(100):
            try:
                self.uid = 0
                self._acting = None

                # All items in the map, inluding agents
                self._items = {}
                # Agents and their current speed.
                # An agent moves when it reaches 0 speed
                self._agents = OrderedDict()
                # All actions available. (agent_id, action_id): function
                self._actions = {}

                min_x, max_x, min_y, max_y = self.map_size
                self.width = random.randint(min_x, max_x)
                self.height = random.randint(min_y, max_y)
                self._map = [[[] for x in range(self.height)]
                             for y in range(self.width)]

                # For estimating best possible reward
                self._approx_reward_map = [[-self.turn_penalty
                                           for x in range(self.height)]
                                          for y in range(self.width)]
                self.__reward_history = dict()
                self.__reward_so_far = 0

                self._reset()
                cornerlocs = [(0, 0),
                              (0, self.height-1),
                              (self.width-1, 0),
                              (self.width-1, self.height-1),
                ]
                for loc in cornerlocs:
                    self._add_item(mi.Corner(location=loc))
                self._step()
                self._accumulate_approximate_rewards()
                self.__approx_best = self._calculate_approximate_reward()

                if self._finished():
                    actor = self.current_agent()
                    self.__reward = self._get_reward(actor)
                    self.__reward_history[actor] = self.__reward_history.get(
                        actor, 0) + self.__reward
                    self.__reward_so_far = self.__reward_history[actor]
                return
            except mazeutils.MazeException:
                logging.exception("Failed to create map because: ")
        raise RuntimeError("Failed to create map after 100 tries! Your map"
                           "size is probably too small")

    def _set_featurizer(self, featurizer):
        '''Helper function for wrappers'''
        self.featurizer = featurizer

    def get_max_bounds(self):
        '''Get maximum width and height across all random initializations'''
        _, max_w, _, max_h = self.map_size
        return max_w, max_h

    @abc.abstractmethod
    def _reset(self):
        '''
        Resets a map to an initial state. Subclass and override this function
        to create new games.
        '''
        pass

    @abc.abstractmethod
    def _finished(self):
        pass

    @abc.abstractmethod
    def _get_reward(self, id):
        reward = -self.turn_penalty
        return reward

    def _accumulate_approximate_rewards(self):
        '''
        Accumulates approximate reward of landing on a square. Used only for
        estimating best possible reward
        '''
        pass

    def _calculate_approximate_reward(self):
        '''
        Accumulates approximate reward of landing on a square. Used only for
        estimating best possible reward
        '''
        return 0

    def _in_bounds(self, location):
        # Checks whether a location is in the maze
        x, y = location
        return 0 <= x < self.width and 0 <= y < self.height

    def _tile_get_block(self, loc, typ):
        for block in self._get_items(loc):
            if isinstance(block, typ):
                return block
        return None

    def _featurize(self, id):
        return self.featurizer.featurize(self, id)

    def _side_info(self):
        '''Override _side_information instead'''
        info = self._side_information()
        for lst in info:
            lst.insert(0, 'INFO')
        return info

    def _side_information(self):
        '''Side information about the game. Shouldn't change too much and
        and encode information about the goals of the game. This list is
        _ordered_, with the information from the superclasses appearing first.

        This is the equivalent of info from mazebase1.0
        '''
        return [['GAME', type(self).__name__]]

    ####################
    # Item functions
    ####################

    def all_possible_features(self):
        '''
        All possible features in the game. Call this to generate a vocabulary
        '''
        if self.__all_possible_features is not None:
            return self.__all_possible_features
        # Circular dependencies
        import mazebase.games as games
        features = set()
        modules = [mi, mi.agents, games]
        for mod in modules:
            for name, cls in mazeutils.all_classes_of(mod):
                features.update(cls.all_features())
        features.update(self.featurizer.all_possible_features(self))
        self.__all_possible_features = list(sorted(features))
        return self.__all_possible_features

    @classmethod
    def all_features(cls):
        '''
        All new features for this game. Usually just the Map Name, and
        you don't need to touch this. If your map implements new features for
        side_info, then define a classmethod with the new features only.
        '''
        return ['GAME', 'INFO', cls.__name__, '']

    def _get_items(self, location):
        # Get item list at a location in the maze, empty if out of buonds
        x, y = location
        if not self._in_bounds(location):
            return []
        return self._map[x][y]

    def _add_item(self, item, id=None):
        assert id is None or isinstance(id, six.string_types) or '|' in id,\
            "Item id must be a string without | characters"
        self.uid += 1
        id = self.game_name + '|' + (str(self.uid) + '|' if id is None else id)
        assert id not in self._items, "Item {0} already in map...".format(id)
        self._items[id] = item

        item.game = self
        item.id = id

        x, y = item.location
        self._map[x][y].append(item)
        return id

    def _move_item(self, id, location):
        nx, ny = location
        if not self._in_bounds(location):
            return
        item = self._items[id]
        x, y = item.location
        self._map[x][y].remove(item)
        self._map[nx][ny].append(item)
        item.location = (nx, ny)

    def _remove_item(self, id):
        item = self._items[id]
        x, y = item.location
        self._map[x][y].remove(item)
        self._items.pop(id)

    ####################
    # Agent functions
    ####################

    @staticmethod
    def all_possible_actions():
        '''
        Returns all possible actions an agent can take
        '''
        actions = set()
        for name, cls in mazeutils.all_classes_of(agents):
            actions.update(cls().actions.keys())
        return list(sorted(actions))

    def actions(self):
        ''' All possible actions for current agent '''
        return sorted([action for agent, action in self._actions.keys()
                if agent == self.current_agent()])

    def current_agent(self):
        '''
        Resets which Agent is doing an action next. We use a countdown model,
        where each agent starts with a speed, and the game ticks down until
        the speed is 0. Then, the agent moves and its speed is reset.
        '''
        if self._acting is None:
            m = min(self._agents.values())
            for k, v in self._agents.items():
                self._agents[k] = v - m
                if v == m:
                    self._acting = k
        if isinstance(self._items[self._acting], agents.NPC):
            self.act(self._items[self._acting].get_npc_action())
            return self.current_agent()
        return self._acting

    def act(self, action):
        ''' Performs an action for current agent '''
        if self._finished():
            return
        actor = self.current_agent()

        # Do nothing if action isn't supported
        def noop():
            logging.debug("Action isn't supported! Passing instead")
        self._actions.get((actor, action), noop)()
        self._step()
        self._agents[actor] = self._items[actor].speed
        self._acting = None

        self.__reward = self._get_reward(actor)
        self.__reward_history[actor] = self.__reward_history.get(actor, 0) + \
            self.__reward
        self.__reward_so_far = self.__reward_history[actor]

    def _add_agent(self, agent, id):
        '''
        Agents are controllable by the player. Non-playing agents should be
        considered items. Agents must have an id to be stable between resets.
        '''
        assert id is not None, "Agent must have an id"
        id = self._add_item(agent, id)
        self._agents[id] = agent.speed
        self._actions.update(dict(((id, k), v) for
                                  k, v in agent.actions.items()))
        return id

    def _step(self):
        '''Hook that is called every time an agent acts'''
        pass


class WithWaterAndBlocksMixin(BaseMazeGame):
    ''' Subcassing this game will generate random blocks and water '''
    __properties = dict(
        waterpct=0.1,
        blockpct=0.1,
        water_penalty=0.2,
    )

    def __init__(self, **kwargs):
        mazeutils.populate_kwargs(self, self.__class__.__properties, kwargs)
        super(WithWaterAndBlocksMixin, self).__init__(**kwargs)

    def _reset(self):
        super(WithWaterAndBlocksMixin, self)._reset()
        creationutils.sprinkle(self, [(mi.Block, self.blockpct),
                                      (mi.Water, self.waterpct)])

    def _get_reward(self, id):
        reward = super(WithWaterAndBlocksMixin, self)._get_reward(id)
        if self._tile_get_block(self._items[id].location, mi.Water) is not None:
            reward += -self.water_penalty
        return reward

    def _accumulate_approximate_rewards(self):
        super(WithWaterAndBlocksMixin, self)._accumulate_approximate_rewards()
        for x, y in product(range(self.width), range(self.height)):
            if self._tile_get_block((x, y), mi.Water) is not None:
                self._approx_reward_map[x][y] += -self.water_penalty


class RewardOnEndMixin(BaseMazeGame):
    '''Subclassing this game will give a final reward that overrides other
    rewards when the episode is finished.

    This should come before any other mixin
    '''
    __properties = dict(
        goal_reward=1,
    )

    def __init__(self, **kwargs):
        mazeutils.populate_kwargs(self, self.__class__.__properties, kwargs)
        super(RewardOnEndMixin, self).__init__(**kwargs)

    def _get_reward(self, id):
        reward = super(RewardOnEndMixin, self)._get_reward(id)
        if self._finished():
            reward = self.goal_reward

        return reward

    def _calculate_approximate_reward(self):
        return super(RewardOnEndMixin, self)._calculate_approximate_reward() + \
            self.goal_reward + self.turn_penalty
        # last turn penalty not counted


class BaseVocabulary(BaseMazeGame):
    '''All the cross-game vocabulary that's needed for most games. Try to add
    as little to this as possible. Presumably games not based off of the
    mazebase gridworld structure won't need this set of vocabulary'''

    def __init__(self, **kwargs):
        self.FEATURE = self.BaseVocabStore()
        super(BaseVocabulary, self).__init__(**kwargs)

    class BaseVocabStore(object):
        '''All non-underscore variables and lists
        will be recorded as a feature'''

        def __init__(self):
            self.GOTO = "GOTO"
            self.IF = "IF"
            self.PUSH = "PUSH"
            self.AVOID = "AVOID"
            self.LEFT = "LEFT"
            self.RIGHT = "RIGHT"
            self.UP = "UP"
            self.DOWN = "DOWN"
            self.SAME = "SAME"
            self.STATE = "STATE"
            self.ALL = "ALL"
            self.SWITCH = "SWITCH"
            self.SAME = "SAME"
            self.ORDERED_OBJ = ['OBJ{0}'.format(i) for i in range(10)]

    @classmethod
    def all_features(cls):
        '''Feature mixins have to call the parent'''
        vocab = cls.BaseVocabStore()
        lst = []
        # flatten lists
        for var_name in dir(vocab):
            if var_name[0] != '_':
                v = getattr(vocab, var_name)
                lst.extend(v) if type(v) == list else lst.append(v)
        return lst + super(BaseVocabulary, cls).all_features()


class AbsoluteLocationVocabulary(mazeutils.AbsoluteLocationUtils, BaseMazeGame):
    '''Featurizer uses absolute locations'''
    def all_possible_features(self):
        return list(sorted(chain(
            super(AbsoluteLocationVocabulary, self).all_possible_features(),
            self._get_abs_loc_features(self)
        )))
